//+------------------------------------------------------------------+
//|                                                     OLAPQRWF.mq5 |
//|                                      Copyright © 2020, Marketeer |
//|                          https://www.mql5.com/en/users/marketeer |
//|               Online Analytical Processing of trading hypercubes |
//+------------------------------------------------------------------+

#property copyright "Copyright © 2020, Marketeer"
#property link      "https://www.mql5.com/ru/articles/7535"
#property version   "1.0"
#property description "Trading Self-Adapting Expert Adviser with OLAP analysis of quotes on-the-fly"

#property strict

enum UPDATEPERIOD
{
  monthly,
  weekly
};

enum CUSTOMFIELD
{
  range,
  count
};

input int BarNumberLookBack = 2880; // BarNumberLookBack (week: 120 H1, month: 480 H1, year: 5760 H1)
input double Threshold = 2.0; // Threshold (PF >= Threshold && PF <= 1/Threshold)
input int Strategy = 0; // Strategy (0 - single bar, 1 - adjacent bars)
input UPDATEPERIOD Update = monthly;
input CUSTOMFIELD CustomField = range;
input int StopLoss = 0;
input double Lot = 0.1;


// forward declaration of our custom class
class CustomQuotesRecord;

// this line plugs our class into default adapter in OLAPQuotes.mqh
#define RECORD_CLASS CustomQuotesRecord


//#include <MT4Bridge/MT4Orders.mqh>
//#include <MT4Bridge/MT4Time.mqh>
#include <OLAP/OLAPQuotes.mqh>
#include <OLAP/SingleNextBarDemo.mqh>


class CustomQuotesRecord: public QuotesRecord
{
  public:
    static string customFieldsTitles[4];

  private:
    static double previousBarRange;
    static double previousSpread;

    int MathSignNonZero(const double x)
    {
      return x >= 0 ? +1 : -1;
    }
    
  public:
    CustomQuotesRecord(): QuotesRecord() {}
    CustomQuotesRecord(const MqlRates &rate): QuotesRecord(rate)
    {
    }
    
    virtual void fillCustomFields() override
    {
      const double newBarRange = get(FIELD_PRICE_RANGE_OC);
      const double spread = get(FIELD_SPREAD);

      if(MathAbs(previousBarRange) > previousSpread)
      {
        double mult = newBarRange * previousBarRange;
        double value = MathSign(mult) * MathAbs(newBarRange);

        // this is an attempt to approximate average losses due to spreads
        value += MathSignNonZero(value) * -1 * MathMax(spread, previousSpread);
        
        set(FIELD_CUSTOM3, value);
        set(FIELD_CUSTOM4, MathSign(mult));
      }
      else
      {
        set(FIELD_CUSTOM3, 0);
        set(FIELD_CUSTOM4, 0);
      }

      set(FIELD_CUSTOM1, MathSign(newBarRange) * (MathAbs(newBarRange) - spread));
      set(FIELD_CUSTOM2, MathSign(newBarRange) * MathSign(MathAbs(newBarRange) - spread));
      
      previousBarRange = newBarRange;
      previousSpread = spread;
    }
    
    virtual string legend(const int index) const override
    {
      if(index >= FIELD_CUSTOM1 && index <= FIELD_CUSTOM4)
      {
        return CustomQuotesRecord::customFieldsTitles[index - FIELD_CUSTOM1];
      }
      
      return QuotesRecord::legend(index);
    }
};

static double CustomQuotesRecord::previousBarRange = 0;
static double CustomQuotesRecord::previousSpread = 0;
static string CustomQuotesRecord::customFieldsTitles[4] =
{
  "Real range (w/o spread)",
  "Real direction counter (w/o spread)",
  "Next bar covariance",
  "Next bar direction counter"
};


const SELECTORS SelectorX = SELECTOR_DAYHOUR;
const ENUM_FIELDS FieldX = FIELD_DATETIME;

const SELECTORS SelectorY = SELECTOR_WEEKDAY;
const ENUM_FIELDS FieldY = FIELD_DATETIME;

const SELECTORS SelectorZ = SELECTOR_NONE;
const ENUM_FIELDS FieldZ = FIELD_NONE;

const SELECTORS _Filter1 = SELECTOR_FILTER;
const ENUM_FIELDS _Filter1Field = FIELD_INDEX;
      int _Filter1value1 = -1; // to be filled with index of first bar to process
const int _Filter1value2 = -1;

const AGGREGATORS _AggregatorType = AGGREGATOR_PROFITFACTOR;
      ENUM_FIELDS _AggregatorField = FIELD_CUSTOM1;
const SORT_BY _SortBy = SORT_BY_NONE;

SELECTORS _selectorArray[4];
ENUM_FIELDS _selectorField[4];



class MyProgress: public Progress
{
  public:
    void progress(const int cursor, const int total) const override
    {
      Comment((string)cursor + " / " + (string)total);
    }
};

MyProgress progress;
datetime lastBar = 0;
bool freshStart = true;


#define PRT(A) Print(#A, " ", (A))

bool isNewBar()
{
  if(lastBar != iTime(_Symbol, 0, 0))
  {
    lastBar = iTime(_Symbol, 0, 0);
    return true;
  }
  return false;
}

int OnInit()
{
  lastBar = 0;
  freshStart = true;

  _selectorArray[0] = SelectorX;
  _selectorArray[1] = SelectorY;
  _selectorArray[2] = SelectorZ;
  _selectorArray[3] = _Filter1;
  _selectorField[0] = FieldX;
  _selectorField[1] = FieldY;
  _selectorField[2] = FieldZ;
  _selectorField[3] = _Filter1Field;
  
  _defaultEngine.setAdapter(&_defaultQuotesAdapter);
  _defaultEngine.setProgress(&progress);
  
  PRT(Bars(_Symbol, _Period));
  
  if(!MQLInfoInteger(MQL_TESTER)) EventSetTimer(1);

  return INIT_SUCCEEDED;
}

#define N_HOURS   24
#define N_DAYS     5
#define AXIS_HOURS 0
#define AXIS_DAYS  1

class MyOLAPStats: public Display
{
  private:
    bool filled;
    double index[][3]; // value, hour, day
    int cursor;

  protected:
    bool saveVector(MetaCube *cube, const int &consts[], const SORT_BY sortby = SORT_BY_NONE)
    {
      PairArray *result = NULL;
      cube.getVector(0, consts, result, sortby);
      if(CheckPointer(result) == POINTER_DYNAMIC)
      {
        const int n = ArraySize(result.array);
        
        if(n == N_HOURS)
        {
          for(int i = 0; i < n; i++)
          {
            index[cursor][0] = result.array[i].value;
            index[cursor][1] = i;
            index[cursor][2] = consts[AXIS_DAYS];
            cursor++;
          }
        }
        
        delete result;
        return n == N_HOURS;
      }
      return false;
    }

  public:
    virtual void display(MetaCube *cube, const SORT_BY sortby = SORT_BY_NONE, const bool identity = false) override
    {
      int consts[];
      const int n = cube.getDimension();
      ArrayResize(consts, n);
      ArrayInitialize(consts, 0);

      filled = false;
      
      ArrayResize(index, N_HOURS * N_DAYS);
      ArrayInitialize(index, 1);
      cursor = 0;

      if(n == 2)
      {
        const int i = AXIS_DAYS;
        int m = cube.getDimensionRange(i); // should be 5 work days
        for(int j = 0; j < m; j++)
        {
          consts[i] = j;
          
          if(!saveVector(cube, consts, sortby)) // 24 hours (values) per current day
          {
            Print("Bad data format");
            return;
          }
          
          consts[i] = 0;
        }
        filled = true;
        ArraySort(index);
        printLines();
      }
      else
      {
        Print("Incorrect cube structure");
      }
    }
    
    void printLines() const
    {
      double temp[3];
      for(int i = 0; i < ArrayRange(index, 0); i++)
      {
        for(int j = 0; j < 3; j++)
        {
          temp[j] = index[i][j];
        }
        _ArrayPrint(temp);
      }
    }
    
    void trade(const double threshold, const double lots, const int strategy = 0)
    {
      const int h = TimeHour(lastBar);
      const int w = TimeDayOfWeek(lastBar) - 1;
    
      int mode = 0;
      
      for(int i = 0; i < N_HOURS * N_DAYS; i++)
      {
        if(index[i][1] == h && index[i][2] == w)
        {
          if(index[i][0] >= threshold)
          {
            mode = +1;
            Print("+ Rule ", i);
            break;
          }
          
          if(index[i][0] <= 1.0 / threshold)
          {
            mode = -1;
            Print("- Rule ", i);
            break;
          }
        }
      }
      
      // pick up existing orders (if any)
      const int direction = CurrentOrderDirection();
      
      if(mode == 0)
      {
        if(direction != 0)
        {
          OrdersCloseAll();
        }
        return;
      }
      
      if(strategy == 0)
      {
        if(direction != 0) // there exist open orders
        {
          if(mode == direction) // keep direction
          {
            return; // existing trade goes on
          }
          OrdersCloseAll();
        }

        const int type = mode > 0 ? OP_BUY : OP_SELL;
        
        const double p = type == OP_BUY ? SymbolInfoDouble(_Symbol, SYMBOL_ASK) : SymbolInfoDouble(_Symbol, SYMBOL_BID);
        const double sl = NormalizeDouble(StopLoss > 0 ? (type == OP_BUY ? p - StopLoss * _Point : p + StopLoss * _Point) : 0, _Digits);
          
        OrderSend(_Symbol, type, Lot, p, 100, sl, 0);
        
      }
      else
      {
        const double previousChange = iOpen(_Symbol, _Period, 0) - iOpen(_Symbol, _Period, 1);
        if(MathAbs(previousChange) <= Ask - Bid /*MathMax(iSpread(_Symbol, _Period, 1), iSpread(_Symbol, _Period, 0)) * _Point*/) return;
      
        const int previousDirection = previousChange > 0 ? +1 : (previousChange < 0 ? -1 : 0);
        
        if(previousDirection == 0) return; // undefined bar
        
        if(direction != 0) // there exist open orders
        {
          if(mode == +1) // bars in same direction
          {
            if(direction == previousDirection) return; // keep existing trade
          }
          else // -1
          {
            if(direction == -previousDirection) return; // keep existing trade
          }
          OrdersCloseAll();
        }
        
        
        const int type = previousDirection * mode > 0 ? OP_BUY : OP_SELL;
        
        const double p = type == OP_BUY ? SymbolInfoDouble(_Symbol, SYMBOL_ASK) : SymbolInfoDouble(_Symbol, SYMBOL_BID);
        const double sl = NormalizeDouble(StopLoss > 0 ? (type == OP_BUY ? p - StopLoss * _Point : p + StopLoss * _Point) : 0, _Digits);
          
        OrderSend(_Symbol, type, Lot, p, 100, sl, 0);
      }
    }
};

MyOLAPStats stats;

bool calcolap()
{
  _Filter1value1 = Bars(_Symbol, _Period) - BarNumberLookBack;
  _AggregatorField = Strategy == 0 ? (ENUM_FIELDS)(FIELD_CUSTOM1 + CustomField) : (ENUM_FIELDS)(FIELD_CUSTOM3 + CustomField);

  _defaultQuotesAdapter.reset();
  const int processed =
  _defaultEngine.process(_selectorArray, _selectorField,
        _AggregatorType, _AggregatorField,
        stats,
        _SortBy,
        _Filter1value1, _Filter1value2);
  
  Comment("");
  
  return processed == BarNumberLookBack;
}

void OnTimer()
{
  EventKillTimer();
  OnTick();
}

void OnTick()
{
  if(!isNewBar()) return;
  
  if(Bars(_Symbol, _Period) < BarNumberLookBack) return;
  
  const int m0 = TimeMonth(iTime(_Symbol, _Period, 0));
  const int w0 = TimeDayOfWeek(iTime(_Symbol, _Period, 0));
  const int m1 = TimeMonth(iTime(_Symbol, _Period, 1));
  const int w1 = TimeDayOfWeek(iTime(_Symbol, _Period, 1));
  
  static bool success = false;
  
  if((Update == monthly && m0 != m1)
  || (Update == weekly && w0 < w1)
  || freshStart)
  {
    success = calcolap();
    freshStart = !success;
  }
  
  if(success)
  {
    stats.trade(Threshold, Lot, Strategy);
  }
  else
  {
    OrdersCloseAll();
  }
  
}

double OnTester()
{
  return TesterStatistics(STAT_SHARPE_RATIO) * TesterStatistics(STAT_TRADES);
}